@@ -23,9 +23,16 @@ import sys
 import sysconfig
 import time
 import warnings
+import logging
 
 from IPython.core.getipython import get_ipython
 
+from spyder_kernels.py3compat import TimeoutError
+from spyder_kernels.comms import CommError
+from spyder_kernels.customize.namespace_manager import NamespaceManager
+
+
+logger = logging.getLogger(__name__)
 
 # We are in Python 2?
 PY2 = sys.version[0] == '2'
@@ -83,9 +90,10 @@ try:
     if os.name == 'nt':
         def encode(u):
             return u.encode('utf8', 'replace')
+
         def execfile(fname, glob=None, loc=None):
             loc = loc if (loc is not None) else glob
-            scripttext = builtins.open(fname).read()+ '\n'
+            scripttext = builtins.open(fname).read() + '\n'
             # compile converts unicode filename to str assuming
             # ascii. Let's do the conversion before calling compile
             if isinstance(fname, unicode):
@@ -94,12 +102,22 @@ try:
                 filename = fname
             exec(compile(scripttext, filename, 'exec'), glob, loc)
     else:
+        def encode(u):
+            return u.encode(sys.getfilesystemencoding())
+
         def execfile(fname, *where):
             if isinstance(fname, unicode):
-                filename = fname.encode(sys.getfilesystemencoding())
+                filename = encode(fname)
             else:
                 filename = fname
             builtins.execfile(filename, *where)
+
+    def maybe_encode(u):
+        if isinstance(u, unicode):
+            return encode(u)
+        else:
+            return u
+
 except ImportError:
     # Python 3
     import builtins
@@ -303,16 +321,34 @@ if os.name == 'nt' and not PY2:
 #==============================================================================
 # Pdb adjustments
 #==============================================================================
-class SpyderPdb(pdb.Pdb):
+class SpyderPdb(pdb.Pdb, object):  # Inherits `object` to call super() in PY2
 
     send_initial_notification = True
     starting = True
 
+    def __init__(self, *args, **kwargs):
+        """Init Pdb."""
+        self.continue_if_has_breakpoints = True
+        super(SpyderPdb, self).__init__(*args, **kwargs)
+
     # --- Methods overriden by us
     def preloop(self):
         """Ask Spyder for breakpoints before the first prompt is created."""
-        if self.starting:
-            get_ipython().kernel._ask_spyder_for_breakpoints()
+        try:
+            _frontend_request(blocking=True).set_debug_state(True)
+            if self.starting:
+                breakpoints = _frontend_request().get_breakpoints()
+                self.set_spyder_breakpoints(breakpoints)
+        except (CommError, TimeoutError):
+            logger.debug("Could not get breakpoints from the frontend.")
+
+    def postloop(self):
+        """Notifies spyder that the loop has ended."""
+        try:
+            _frontend_request(blocking=True).set_debug_state(False)
+        except (CommError, TimeoutError):
+            logger.debug("Could not send debugging state to the frontend.")
+        super(SpyderPdb, self).postloop()
 
     # --- Methods defined by us
     def set_spyder_breakpoints(self, breakpoints):
@@ -347,10 +383,18 @@ class SpyderPdb(pdb.Pdb):
             # Do 'continue' if the first breakpoint is *not* placed
             # where the debugger is going to land.
             # Fixes issue 4681
-            if breaks and lineno != breaks[0] and osp.isfile(fname):
-                get_ipython().kernel.pdb_continue()
+            if (self.continue_if_has_breakpoints and
+                    breaks and
+                    lineno < breaks[0] and
+                    osp.isfile(fname)):
+                try:
+                    get_ipython().kernel.pdb_continue()
+                except (CommError, TimeoutError):
+                    logger.debug(
+                        "Could not send a Pdb continue call to the frontend.")
 
     def notify_spyder(self, frame):
+        """Send kernel state to the frontend."""
         if not frame:
             return
 
@@ -374,7 +418,10 @@ class SpyderPdb(pdb.Pdb):
         # Publish Pdb state so we can update the Variable Explorer
         # and the Editor on the Spyder side
         kernel._pdb_step = step
-        kernel.publish_pdb_state()
+        try:
+            kernel.publish_pdb_state()
+        except (CommError, TimeoutError):
+            logger.debug("Could not send Pdb state to the frontend.")
 
 
 pdb.Pdb = SpyderPdb
@@ -436,7 +483,7 @@ def user_return(self, frame, return_value):
 @monkeypatch_method(pdb.Pdb, 'Pdb')
 def interaction(self, frame, traceback):
     if frame is not None and "spydercustomize.py" in frame.f_code.co_filename:
-        self.run('exit')
+        self.onecmd('exit')
     else:
         self.setup(frame, traceback)
         if self.send_initial_notification:
@@ -473,7 +520,8 @@ def reset(self):
 #     specific behaviour desired?)
 @monkeypatch_method(pdb.Pdb, 'Pdb')
 def postcmd(self, stop, line):
-    if "_set_spyder_breakpoints" not in line:
+    if ("_set_spyder_breakpoints" not in line and
+            '!get_ipython().kernel' not in line):
         self.notify_spyder(self.curframe)
     return self._old_Pdb_postcmd(stop, line)
 
@@ -777,19 +825,64 @@ if "SPYDER_EXCEPTHOOK" in os.environ:
 # ==============================================================================
 # runfile and debugfile commands
 # ==============================================================================
-def _get_globals():
-    """Return current namespace"""
-    ipython_shell = get_ipython()
-    return ipython_shell.user_ns
+def _frontend_request(blocking=True):
+    """
+    Send a request to the frontend.
+
+    If blocking is true, The return value will be returned.
+    """
+    if not get_ipython().kernel.frontend_comm.is_open():
+        raise CommError("Can't make a request to a closed comm")
+    # Get a reply from the last frontend to have sent a message
+    return get_ipython().kernel.frontend_call(
+        blocking=blocking, broadcast=False)
 
 
-def runfile(filename, args=None, wdir=None, namespace=None, post_mortem=False):
+def get_current_file_name():
+    """Get the current file name."""
+    try:
+        return _frontend_request().current_filename()
+    except Exception:
+        _print("This command failed to be executed because an error occurred"
+               " while trying to get the current file name from Spyder's"
+               " editor. The error was:\n\n")
+        get_ipython().showtraceback(exception_only=True)
+        return None
+
+
+def get_debugger(filename):
+    """Get a debugger for a given filename."""
+    debugger = pdb.Pdb()
+    filename = debugger.canonic(filename)
+    debugger._wait_for_mainpyfile = 1
+    debugger.mainpyfile = filename
+    debugger._user_requested_quit = 0
+    if os.name == 'nt':
+        filename = filename.replace('\\', '/')
+    return debugger, filename
+
+
+def runfile(filename=None, args=None, wdir=None, namespace=None,
+            post_mortem=False, is_pdb=False, current_namespace=False):
     """
     Run filename
     args: command line arguments (string)
     wdir: working directory
+    namespace: namespace for execution
     post_mortem: boolean, whether to enter post-mortem mode on error
+    current_namespace: if true, run the file in the current namespace
     """
+    ipython_shell = get_ipython()
+    if filename is None:
+        filename = get_current_file_name()
+        if filename is None:
+            return
+    try:
+        # Save the open files
+        _frontend_request().save_files()
+    except Exception:
+        logger.debug("Could not save files before executing.")
+
     try:
         filename = filename.decode('utf-8')
     except (UnicodeError, TypeError, AttributeError):
@@ -800,46 +893,71 @@ def runfile(filename, args=None, wdir=None, namespace=None, post_mortem=False):
         __umr__.run()
     if args is not None and not isinstance(args, basestring):
         raise TypeError("expected a character buffer object")
-    if namespace is None:
-        namespace = _get_globals()
-    namespace['__file__'] = filename
-    sys.argv = [filename]
-    if args is not None:
-        for arg in shlex.split(args):
-            sys.argv.append(arg)
-    if wdir is not None:
-        try:
-            wdir = wdir.decode('utf-8')
-        except (UnicodeError, TypeError, AttributeError):
-            # UnicodeError, TypeError --> eventually raised in Python 2
-            # AttributeError --> systematically raised in Python 3
-            pass
-        os.chdir(wdir)
-    if post_mortem:
-        set_post_mortem()
-    if __umr__.has_cython:
-        # Cython files
-        with io.open(filename, encoding='utf-8') as f:
-            ipython_shell = get_ipython()
-            ipython_shell.run_cell_magic('cython', '', f.read())
-    else:
-        execfile(filename, namespace)
 
-    clear_post_mortem()
-    sys.argv = ['']
+    with NamespaceManager(filename, namespace, current_namespace) as namespace:
+        sys.argv = [filename]
+        if args is not None:
+            for arg in shlex.split(args):
+                sys.argv.append(arg)
+        if wdir is not None:
+            try:
+                wdir = wdir.decode('utf-8')
+            except (UnicodeError, TypeError, AttributeError):
+                # UnicodeError, TypeError --> eventually raised in Python 2
+                # AttributeError --> systematically raised in Python 3
+                pass
+            os.chdir(wdir)
+        if post_mortem:
+            set_post_mortem()
 
-    # Avoid error when running `%reset -f` programmatically
-    # See issue spyder-ide/spyder-kernels#91
-    try:
-        namespace.pop('__file__')
-    except KeyError:
-        pass
+        if __umr__.has_cython:
+            # Cython files
+            with io.open(filename, encoding='utf-8') as f:
+                ipython_shell.run_cell_magic('cython', '', f.read())
+        else:
+            try:
+                execfile(filename, namespace)
+            except SystemExit as status:
+                # ignore exit(0)
+                if status.code:
+                    ipython_shell.showtraceback(exception_only=True)
+            except BaseException as error:
+                if isinstance(error, bdb.BdbQuit) and is_pdb:
+                    # Ignore BdbQuit if we are debugging, as it is expected.
+                    pass
+                else:
+                    # We ignore the call to execfile and exec
+                    ipython_shell.showtraceback(tb_offset=2)
+
+        clear_post_mortem()
+        sys.argv = ['']
 
 
 builtins.runfile = runfile
 
 
-def runcell(cellname, filename):
+def debugfile(filename=None, args=None, wdir=None, post_mortem=False,
+              current_namespace=False):
+    """
+    Debug filename
+    args: command line arguments (string)
+    wdir: working directory
+    post_mortem: boolean, included for compatiblity with runfile
+    """
+    if filename is None:
+        filename = get_current_file_name()
+        if filename is None:
+            return
+    debugger, filename = get_debugger(filename)
+    debugger.run("runfile(%r, args=%r, wdir=%r, is_pdb=True, "
+                          "current_namespace=%r)" % (
+        filename, args, wdir, current_namespace))
+
+
+builtins.debugfile = debugfile
+
+
+def runcell(cellname, filename=None, is_pdb=False):
     """
     Run a code cell from an editor as a file.
 
@@ -849,12 +967,15 @@ def runcell(cellname, filename):
 
     Parameters
     ----------
-    cellname : str
-        Used as a reference in the history log of which
-        cell was run with the fuction. This variable is not used.
+    cellname : str or int
+        Cell name or index.
     filename : str
         Needed to allow for proper traceback links.
     """
+    if filename is None:
+        filename = get_current_file_name()
+        if filename is None:
+            return
     try:
         filename = filename.decode('utf-8')
     except (UnicodeError, TypeError, AttributeError):
@@ -862,46 +983,85 @@ def runcell(cellname, filename):
         # AttributeError --> systematically raised in Python 3
         pass
     ipython_shell = get_ipython()
-    namespace = _get_globals()
-    namespace['__file__'] = filename
     try:
-        cell_code = ipython_shell.cell_code
-    except AttributeError:
-        _print("--Run Cell Error--\n"
-               "Please use only through Spyder's Editor; "
-               "shouldn't be called manually from the console")
+        # Get code from spyder
+        cell_code = _frontend_request().run_cell(cellname, filename)
+    except Exception:
+        _print("This command failed to be executed because an error occurred"
+               " while trying to get the cell code from Spyder's"
+               " editor. The error was:\n\n")
+        get_ipython().showtraceback(exception_only=True)
+        return
+
+    if not cell_code:
+        # Nothing to execute
         return
 
     # Trigger `post_execute` to exit the additional pre-execution.
     # See Spyder PR #7310.
     ipython_shell.events.trigger('post_execute')
 
-    ipython_shell.run_cell(cell_code)
-    namespace.pop('__file__')
-    del ipython_shell.cell_code
+    with NamespaceManager(filename, current_namespace=True) as namespace:
+        try:
+            if PY2:
+                filename = maybe_encode(filename)
+            exec(compile(cell_code, filename, 'exec'), namespace)
+        except SystemExit as status:
+            # ignore exit(0)
+            if status.code:
+                ipython_shell.showtraceback(exception_only=True)
+        except BaseException as error:
+            if isinstance(error, bdb.BdbQuit) and is_pdb:
+                # Ignore BdbQuit if we are debugging, as it is expected.
+                pass
+            else:
+                # We ignore the call to exec
+                ipython_shell.showtraceback(tb_offset=1)
 
 
 builtins.runcell = runcell
 
 
-def debugfile(filename, args=None, wdir=None, post_mortem=False):
-    """
-    Debug filename
-    args: command line arguments (string)
-    wdir: working directory
-    post_mortem: boolean, included for compatiblity with runfile
-    """
-    debugger = pdb.Pdb()
-    filename = debugger.canonic(filename)
-    debugger._wait_for_mainpyfile = 1
-    debugger.mainpyfile = filename
-    debugger._user_requested_quit = 0
-    if os.name == 'nt':
-        filename = filename.replace('\\', '/')
-    debugger.run("runfile(%r, args=%r, wdir=%r)" % (filename, args, wdir))
+def debugcell(cellname, filename=None):
+    """Debug a cell."""
+    if filename is None:
+        filename = get_current_file_name()
+        if filename is None:
+            return
+
+    debugger, filename = get_debugger(filename)
+    # The breakpoint might not be in the cell
+    debugger.continue_if_has_breakpoints = False
+    debugger.run("runcell({}, {}, is_pdb=True)".format(
+        repr(cellname), repr(filename)))
 
 
-builtins.debugfile = debugfile
+builtins.debugcell = debugcell
+
+
+def cell_count(filename=None):
+    """
+    Get the number of cells in a file.
+
+    Parameters
+    ----------
+    filename : str
+        The file to get the cells from. If None, the currently opened file.
+    """
+    if filename is None:
+        filename = get_current_file_name()
+        if filename is None:
+            raise RuntimeError('Could not get cell count from frontend.')
+    try:
+        # Get code from spyder
+        cell_count = _frontend_request().cell_count(filename)
+        return cell_count
+    except Exception:
+        etype, error, tb = sys.exc_info()
+        raise etype(error)
+
+
+builtins.cell_count = cell_count
 
 
 #==============================================================================
